/** @file
===========================================================================
MODULE				: CCropFilter

DESCRIPTIONS		: Crop Filter for RGB24 and RGB32 Media

REVISION HISTORY	: Version 1.0
					27/08/2008: RG Fixed boolean logic bug in CheckTransform which resulted in the VMR9 rendering an incorrect image
					27/08/2008: RG Adjusted source and target rect in GetMediaType


COPYRIGHT			: (c)CSIR, Meraka Institute 2007 all rights reserved

RESTRICTIONS		: The information/data/code contained within this file is 
					  the property of CSIR, Meraka Institute and has been 
					  classified as PUBLIC.
===========================================================================
*/
#include "CropFilter.h"

#include <DirectShow/Common/datastructures.h>

#include <Utils/ImageUtils/PicCropperRGB24Impl.h>
#include <Utils/ImageUtils/PicCropperRGB32Impl.h>

CCropFilter::CCropFilter()
: CCustomBaseFilter(NAME("Meraka RTVC Crop Filter"), 0, CLSID_CropFilter),
	m_pCropper(NULL),
	m_nBytesPerPixel(BYTES_PER_PIXEL_RGB24),
	m_nLeftCrop(0),
	m_nRightCrop(0),
	m_nTopCrop(0),
	m_nBottomCrop(0),
	m_nTopCropRatio(1),
	m_nBottomCropRatio(2),
	m_nLeftCropRatio(1),
	m_nRightCropRatio(1)
{
	//Call the initialise input method to load all acceptable input types for this filter
	InitialiseInputTypes();
	//Default values
	m_nOutWidth = 0;
	m_nOutHeight = 0;
}

CCropFilter::~CCropFilter()
{
	if (m_pCropper)
	{
		delete m_pCropper;
		m_pCropper = NULL;
	}
}

CUnknown * WINAPI CCropFilter::CreateInstance( LPUNKNOWN pUnk, HRESULT *pHr )
{
	CCropFilter *pFilter = new CCropFilter();
	if (pFilter== NULL) 
	{
		*pHr = E_OUTOFMEMORY;
	}
	return pFilter;
}


void CCropFilter::InitialiseInputTypes()
{
	AddInputType(&MEDIATYPE_Video, &MEDIASUBTYPE_RGB24, &FORMAT_VideoInfo);
	AddInputType(&MEDIATYPE_Video, &MEDIASUBTYPE_RGB32, &FORMAT_VideoInfo);
}

HRESULT CCropFilter::SetMediaType( PIN_DIRECTION direction, const CMediaType *pmt )
{
	HRESULT hr = CCustomBaseFilter::SetMediaType(direction, pmt);
	if (direction == PINDIR_INPUT)
	{
		//Set defaults and make sure that the target dimensions are valid
		if ((m_nOutWidth == 0)||(m_nOutWidth > m_nInWidth))
		{
			m_nOutWidth = m_nInWidth;
		}
		if ((m_nOutHeight == 0)||(m_nOutHeight > m_nInHeight))
		{
			m_nOutHeight = m_nInHeight;
		}
		//Determine whether we are connected to a RGB24 or 32 source
		if (pmt->majortype == MEDIATYPE_Video)
		{
			//The converter might already exist if the filter has been connected previously
			if (m_pCropper)
			{
				delete m_pCropper;
				m_pCropper = NULL;
			}
			if (pmt->subtype == MEDIASUBTYPE_RGB24)
			{
				m_pCropper = new PicCropperRGB24Impl();
				m_nBytesPerPixel = BYTES_PER_PIXEL_RGB24;
			}
			else if (pmt->subtype == MEDIASUBTYPE_RGB32)
			{
				m_pCropper = new PicCropperRGB32Impl();
				m_nBytesPerPixel = BYTES_PER_PIXEL_RGB32;
			}
		}
		RecalculateFilterParameters();
	}
	return hr;
}

HRESULT CCropFilter::GetMediaType( int iPosition, CMediaType *pMediaType )
{
	if (iPosition < 0)
	{
		return E_INVALIDARG;
	}
	else if (iPosition == 0)
	{
		// Get the input pin's media type and return this as the output media type - we want to retain
		// all the information about the image
		HRESULT hr = m_pInput->ConnectionMediaType(pMediaType);
		if (FAILED(hr))
		{
			return hr;
		}
		// Get the bitmap info header and adapt the cropped 
		//make sure that it's a video info header
		ASSERT(pMediaType->formattype == FORMAT_VideoInfo);
		VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER*)pMediaType->pbFormat;
		//Now we need to calculate the size of the output image
		BITMAPINFOHEADER* pBi = &(pVih->bmiHeader);
		
		pBi->biHeight = m_nOutHeight;
		ASSERT(pBi->biHeight > 0);

		pBi->biWidth = m_nOutWidth;
		ASSERT(pBi->biWidth > 0);
		pBi->biSizeImage = pBi->biWidth * pBi->biHeight * m_nBytesPerPixel;

		pVih->rcSource.top = 0;
		pVih->rcSource.left = 0;
		pVih->rcSource.right = pBi->biWidth;
		pVih->rcSource.bottom = pBi->biHeight;

		pVih->rcTarget.top = 0;
		pVih->rcTarget.left = 0;
		pVih->rcTarget.right = pBi->biWidth;
		pVih->rcTarget.bottom = pBi->biHeight;

		pMediaType->lSampleSize = pBi->biSizeImage;

		return S_OK;
	}
	return VFW_S_NO_MORE_ITEMS;
}

HRESULT CCropFilter::DecideBufferSize( IMemAllocator *pAlloc, ALLOCATOR_PROPERTIES *pProp )
{
	pProp->cbBuffer = m_nOutPixels * m_nBytesPerPixel;

	if (pProp->cbAlign == 0)
	{
		pProp->cbAlign = 1;
	}
	if (pProp->cBuffers == 0)
	{
		pProp->cBuffers = 1;
	}
	
	// Set allocator properties.
	ALLOCATOR_PROPERTIES Actual;
	HRESULT hr = pAlloc->SetProperties(pProp, &Actual);
	if (FAILED(hr)) 
	{
		return hr;
	}
	// Even when it succeeds, check the actual result.
	if (pProp->cbBuffer > Actual.cbBuffer) 
	{
		return E_FAIL;
	}
	return S_OK;
}

HRESULT CCropFilter::CheckTransform( const CMediaType *mtIn, const CMediaType *mtOut )
{
	//Make sure the input and output types are related
	if (mtOut->majortype != MEDIATYPE_Video)
	{
		return VFW_E_TYPE_NOT_ACCEPTED;
	}
	// RG: 27/08/2008 Bug FIX: Resulted in incorrect image in VMR9
	if (!
		((mtIn->subtype == MEDIASUBTYPE_RGB24)&&(mtOut->subtype == MEDIASUBTYPE_RGB24))
		||
		((mtIn->subtype == MEDIASUBTYPE_RGB32)&&(mtOut->subtype == MEDIASUBTYPE_RGB32))
		)
	{
		return VFW_E_TYPE_NOT_ACCEPTED;
	}
	if (mtOut->formattype != FORMAT_VideoInfo)
	{
		return VFW_E_TYPE_NOT_ACCEPTED;
	}
	return S_OK;
}


STDMETHODIMP CCropFilter::GetParameter( const char* type, char* szValue, int& nResultLength )
{
	char *p = (char *)type;
	int len	= (int)strlen(p);

	if( _strnicmp(p, TARGET_WIDTH,len) == 0 )
		_itoa(m_nOutWidth,(char *)szValue,10);
	else if( _strnicmp(p, TARGET_HEIGHT, len) == 0 )
		_itoa(m_nOutHeight,(char *)szValue,10);
	else if( _strnicmp(p, LEFT_CROP_RATIO, len) == 0 )
		_itoa(m_nLeftCropRatio,(char *)szValue,10);
	else if( _strnicmp(p, RIGHT_CROP_RATIO, len) == 0 )
		_itoa(m_nRightCropRatio,(char *)szValue,10);
	else if( _strnicmp(p, TOP_CROP_RATIO, len) == 0 )
		_itoa(m_nTopCropRatio,(char *)szValue,10);
	else if( _strnicmp(p, BOTTOM_CROP_RATIO, len) == 0 )
		_itoa(m_nBottomCropRatio,(char *)szValue,10);
	else
	{
		//_errorStr = "Cropper: Read parameter not supported";
		nResultLength = -1;
		return S_FALSE;
	}//end else...

	nResultLength = (int)strlen(szValue);
	return S_OK;
}

STDMETHODIMP CCropFilter::SetParameter( const char* type, const char* value )
{
	// For now, one cannot set any parameters once the output has been connected -> this will affect the buffer size
	// m_pOutput only gets created in the Trnasform filter once the input pin has been connected - need to check pointer
	if (m_pOutput)
	{
		if (m_pOutput->IsConnected())
		{
			DbgLog((LOG_TRACE, 0, TEXT("Unable to set parameter: output pins is already connected - no dynamic reconnection support")));
			return E_FAIL;
		}
	}

	char *p = (char *)type;
	char *v = (char *)value;
	int  len	= (int)strlen(p);

	if( _strnicmp(p, TARGET_WIDTH, len) == 0 )
	{
		int nWidth = (atoi(v));
		// Make sure the value is valid
		return SetDimensionIfValid(nWidth, m_nInWidth, m_nOutWidth);
	}
	else if( _strnicmp(p, TARGET_HEIGHT, len) == 0 )
	{
		int nHeight = (atoi(v));
		return SetDimensionIfValid(nHeight, m_nInHeight, m_nOutHeight);
		//if (nHeight > 0)
		//{
		//	// If the in height has been set, i.e filter is connected
		//	if (m_nInHeight > 0)
		//		if (nHeight <= m_nInHeight)
		//			m_nOutHeight = nHeight;
		//		else
		//			return E_FAIL;
		//	else
		//		m_nOutHeight = nHeight;
		//}
		//else
		//{
		//	return E_FAIL;
		//}
	}
	else if( _strnicmp(p, LEFT_CROP_RATIO, len) == 0 )
	{
		int nRatio = (atoi(v));
		return SetRatioIfValid(nRatio, m_nLeftCropRatio, m_nRightCropRatio);
	}
	else if( _strnicmp(p, RIGHT_CROP_RATIO, len) == 0 )
	{
		int nRatio = (atoi(v));
		return SetRatioIfValid(nRatio, m_nRightCropRatio, m_nLeftCropRatio);
	}
	else if( _strnicmp(p, TOP_CROP_RATIO, len) == 0 )
	{
		int nRatio = (atoi(v));
		return SetRatioIfValid(nRatio, m_nTopCropRatio, m_nBottomCropRatio);
	}
	else if( _strnicmp(p, BOTTOM_CROP_RATIO, len) == 0 )
	{
		int nRatio = (atoi(v));
		return SetRatioIfValid(nRatio, m_nBottomCropRatio, m_nTopCropRatio);
	}
	else
	{
		//_errorStr = "Cropper: Write parameter not supported";
		return E_FAIL;
	}//end else...
	return RecalculateFilterParameters();
	//return S_OK;
}

DWORD CCropFilter::ApplyTransform( BYTE* pBufferIn, BYTE* pBufferOut )
{
	int nTotalSize = 0;
	//make sure we were able to initialise our converter
	if (m_pCropper)
	{
		//Call cropping conversion code
		m_pCropper->SetInDimensions(m_nInWidth, m_nInHeight);
		m_pCropper->SetOutDimensions(m_nOutWidth, m_nOutHeight);
		m_pCropper->SetCrop(m_nLeftCrop, m_nRightCrop, m_nTopCrop, m_nBottomCrop);
		m_pCropper->Crop((void*)pBufferIn, (void*)pBufferOut);
		//Total pixels is now 4 times larger
		nTotalSize = m_nOutWidth * m_nOutHeight * m_nBytesPerPixel;
	}
	else
	{
		DbgLog((LOG_TRACE, 0, TEXT("TestCropper: Cropper is not initialised - unable to transform")));
	}
	return nTotalSize;
}


HRESULT CCropFilter::SetDimensionIfValid(int nNewValue, int& nInValue, int& nOutValue )
{
	if (nNewValue > 0)
	{
		//Check if filter has been connected yet
		if (nInValue > 0)
		{
			// make sure value is valid
			if (nNewValue <= nInValue)
			{
				// Set out value
				nOutValue = nNewValue;
				return S_OK;
			}
			else
			{
				return E_FAIL;
			}
		}
		else
		{
			// accept the value for now - will be reset if invalid when the input is connected
			nOutValue = nNewValue;
			return S_OK;
		}
	}
	else 
	{
		return E_FAIL;
	}
}

HRESULT CCropFilter::RecalculateFilterParameters()
{
	//Make sure the input has been connected
	if ((m_nInHeight > 0) && (m_nInWidth > 0))
	//Now calculate all the cropper offsets
	{
		int nOffsetHeight = m_nInHeight - m_nOutHeight;
		if (nOffsetHeight >= 0)
		{
			ASSERT(((m_nTopCropRatio > 0) || (m_nBottomCropRatio > 0)));
			m_nBottomCrop = (m_nBottomCropRatio * nOffsetHeight) / (m_nBottomCropRatio + m_nTopCropRatio);
			//Now calculate the top relative to the bottom in case of floating point errors
			m_nTopCrop = nOffsetHeight - m_nBottomCrop;
		}
		else
		{
			return E_FAIL;
		}

		int nOffsetWidth = m_nInWidth - m_nOutWidth;
		if (nOffsetWidth >= 0)
		{
			ASSERT(((m_nLeftCropRatio > 0) || (m_nRightCropRatio > 0)));
			m_nLeftCrop = (m_nLeftCropRatio * nOffsetWidth) / (m_nLeftCropRatio + m_nRightCropRatio);
			//Now calculate the top relative to the bottom in case of floating point errors
			m_nRightCrop = nOffsetWidth - m_nLeftCrop;
		}
		else
		{
			return E_FAIL;
		}
	}
	return S_OK;
}

HRESULT CCropFilter::SetRatioIfValid( int nNewValue, int& nRatio1, int& nRatio2 )
{
	if (nNewValue > 0)
	{
		nRatio1 = nNewValue;
		return S_OK;
	}
	else if (nNewValue == 0)
	{
		// Only set the ratio if the opposite is non zero
		if (nRatio2 != 0)
		{
			nRatio1 = nNewValue;
			return S_OK;
		}
		else
		{
			// Both ratio's can't be zero
			return E_FAIL;
		}
	}
	else
	{
		// negative value
		return E_FAIL;
	}
}