/** @file
===========================================================================
MODULE				: CCustomBaseFilter

DESCRIPTIONS		: Scale Filter for RGB24 Media
						Parameters are settable via the ISettingsInterface COM interface and can only be set before the 
						downstream filter is connected.


REVISION HISTORY	: Version 1.1
					  1) 15/02/2008 RG: Renamed variable names and changed SettingsInterface signatures to be more intuitive


COPYRIGHT			: (c)CSIR, Meraka Institute 2007 all rights reserved

RESTRICTIONS		: The information/data/code contained within this file is 
						the property of CSIR, Meraka Institute and has been 
						classified as PUBLIC.
===========================================================================
*/

#pragma once

#include <DirectShow/Common/CustomBaseFilter.h>

// {60E0821D-9E6C-48f4-8E1C-1B25E67488F8}
static const GUID CLSID_ScaleFilter = 
{ 0x60e0821d, 0x9e6c, 0x48f4, { 0x8e, 0x1c, 0x1b, 0x25, 0xe6, 0x74, 0x88, 0xf8 } };

#ifdef _BUILD_FOR_SHORT
typedef short yuvType;
#else
typedef signed char yuvType ;
#endif

// Forward declarations
class PicScalerBase;


class CScaleFilter : public CCustomBaseFilter
{
public:
	/// Constructor
	CScaleFilter();
	/// Destructor
	~CScaleFilter();

	/// Static object-creation method (for the class factory)
	static CUnknown * WINAPI CreateInstance(LPUNKNOWN pUnk, HRESULT *pHr); 

	/**
	* Overriding this so that we can set whether this is an RGB24 or an RGB32 Filter
	*/
	HRESULT SetMediaType(PIN_DIRECTION direction, const CMediaType *pmt);

	/**
	* Used for Media Type Negotiation 
	* Returns an HRESULT value. Possible values include those shown in the following table.
	* <table border="0" cols="2"><tr valign="top"><td><b>Value</b></td><td><b>Description</b></td></TR><TR><TD>S_OK</TD><TD>Success</TD></TR><TR><TD>VFW_S_NO_MORE_ITEMS</TD><TD>Index out of range</TD></TR><TR><TD>E_INVALIDARG</TD><TD>Index less than zero</TD></TR></TABLE>
	* The output pin's CTransformOutputPin::GetMediaType method calls this method. The derived class must implement this method. For more information, see CBasePin::GetMediaType.
	* To use custom media types, the media type can be manipulated in this method.
	*/
	HRESULT GetMediaType(int iPosition, CMediaType *pMediaType);

	/// Buffer Allocation
	/**
	* The output pin's CTransformOutputPin::DecideBufferSize method calls this method. The derived class must implement this method. For more information, see CBaseOutputPin::DecideBufferSize. 
	* @param pAlloc Pointer to the IMemAllocator interface on the output pin's allocator.
	* @param pProp Pointer to an ALLOCATOR_PROPERTIES structure that contains buffer requirements from the downstream input pin.
	* @return Value: Returns S_OK or another HRESULT value.
	*/
	HRESULT DecideBufferSize(IMemAllocator *pAlloc, ALLOCATOR_PROPERTIES *pProp);

	/**
	* The CheckTransform method checks whether an input media type is compatible with an output media type.
	* <table border="0" cols="2"> <tr valign="top"> <td  width="50%"><b>Value</b></td> <td width="50%"><b>Description</b></td> </tr> <tr valign="top"> <td width="50%">S_OK</td> <td width="50%">The media types are compatible.</td> </tr> <tr valign="top"> <td width="50%">VFW_E_TYPE_NOT_ACCEPTED</td> <td width="50%">The media types are not compatible.</td> </tr> </table>
	*/
	HRESULT CheckTransform(const CMediaType *mtIn, const CMediaType *mtOut);

	/// Interface methods
	///Overridden from ISettingsInterface
	STDMETHODIMP GetParameter( const char* szParameterName, char* szResult, int& nResultLength );
	STDMETHODIMP SetParameter( const char* type, const char* szValue);

	/// Overridden from CTMBaseFilter
	virtual void InitialiseInputTypes();
protected:


private:
	/**
	* This method converts the input buffer from RGB24 | 32 to YUV420P
	* @param pSource The source buffer
	* @param pDest The destination buffer
	*/
	DWORD ApplyTransform(BYTE* pBufferIn, BYTE* pBufferOut);
	
	HRESULT RecalculateFilterParameters();

	/// Pointer to our picture scaler
	PicScalerBase* m_pScaler;
	/// Stores bytes per needed to store pixel according to media type
	double m_nBytesPerPixel;
};